/* SerialToVM: Interface between RS-232 and SEGA Dreamcast VM */
/* Implements several systemdependant low-lovel routines */
/* This file is for a PC DOS parallel port interface */

/*******************************************************************
This parallel port driver written by John Maushammer
http://www.tidalwave.net/~maushammer/vmu.html


********************************************************************/


/*******************************************************************
Copyright (C) 2000 by Soeren Gust, sgust@ithh.infoserv.de

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

You can get the latest Version from http://soeren.infoserv.de
********************************************************************/

#ifndef _WIN32
#include <termios.h>
#include <unistd.h>
#include <sys/time.h>
#endif
#include <stdio.h>
#include <fcntl.h>

#include "lowlevel_dosparallel.h"
#include "misc.h"

/* open and initialise the serial port "serial" described by "serialname",
   return 0 on error, 1 on success */
int openserial(char *serialname, serial_t *serial)
{
	struct termios serialio;

	/* open file */
	*serial = open(serialname, O_RDWR | O_SYNC);
	/* test for error */
	if (*serial < 0) return 0;
	/* get terminal attributes */
	if (tcgetattr(*serial, &serialio) != 0) return 0;
	/* make sure nobody changes any of the bytes as they go in and out */
	cfmakeraw(&serialio);
	/* 8 bits, receive, ignore modem control (DCD), enable RTS/CTS, drop DTR on close */
	serialio.c_cflag |= CS8 | CREAD | CLOCAL | CRTSCTS | HUPCL;
	/* 19200 bps in and out */
	if (cfsetospeed(&serialio, B19200) != 0) return 0;
	if (cfsetispeed(&serialio, B19200) != 0) return 0;
	/* set new parameters now */
	if (tcsetattr(*serial, TCSANOW, &serialio) != 0) return 0;
	return 1; /* everything OK */
}

/* close serial port "serial",
   return 0 on error, 1 on success */
int closeserial(serial_t *serial)
{
	int i;

	/* close serial port */
	i = close(*serial);
	/* destroy descriptor */
	*serial = 0;
	/* deliver return code */
	if (i == -1) return 0; else return 1;
}

/* read one byte from the serial port "serial", give up after "timeout" seconds,
   return -1 on error, -2 on timeout, the byte received cast to an int on success */
int readserial(serial_t *serial, int timeout)
{
	int i;
	fd_set readfds;
	struct timeval timeval;
	byte c;

	/* fill in timeout */
	timeval.tv_sec = timeout;
	timeval.tv_usec = 0;
	/* fill in file descriptor */
	FD_ZERO(&readfds);
	FD_SET(*serial, &readfds);
	/* call select */
	i = select(*serial + 1, &readfds, NULL, NULL, &timeval);
	/* error in select */
	if (i == -1) return -1;
	else {
		/* check if really ok to read */
		if (FD_ISSET(*serial, &readfds)) {
			/* read byte */
			i = read(*serial, &c, 1);
			/* check for error */
			if (i == 1) return (int) c; /* read 1 byte */
			else return -1;
		}
		/* select returned, but nothing to read: timeout */
		else return -2;
	}
}

/* wite one byte c to the serial port "serial", give up after "timeout" seconds,
   return -1 on error, -2 on timeout, 0 on success */
int writeserial(serial_t *serial, byte c, int timeout)
{
	int result;
	fd_set writefds;
	struct timeval timeval;

	/* fill in timeout */
	timeval.tv_sec = timeout;
	timeval.tv_usec = 0;
	/* fill in file descriptor */
	FD_ZERO(&writefds);
	FD_SET(*serial, &writefds);
	/* call select */
	result = select(*serial + 1, NULL, &writefds, NULL, &timeval);
	/* error in select */
	if (result == -1) return -1;
	else {
		/* check if really ok to write */
		if (FD_ISSET(*serial, &writefds)) {
			/* write byte */
			result = write(*serial, &c, 1);
			/* check for error */
			if (result == 1) return 0; /* 1 byte written, ok */
			else return -1;
		}
		/* select returned, but can't write: timeout */
		else return -2;
	}
}
