#include <windows.h>
#include <winbase.h>
#include <time.h>
/* SerialToVM: Interface between RS-232 and SEGA Dreamcast VM */
/* Implements several systemdependant low-lovel routines */
/* This file is for a PC DOS parallel port interface */

/* I got this when I didn't have the delayed 8-bit data out:
 * Communications log: R:11 S:e0 R:02 R:33 R:0a R:ff S:0c R:10 S:e0 R:0b R:00 R:ff (one more $FF seen on scope but not rx'ed)
 *
 * pin 4 is driven with a byte delayed version of what we RX'd not what
 * we sent.
 */

/*******************************************************************
This parallel port driver written by John Maushammer
http://www.tidalwave.net/~maushammer/vmu.html

** ALPHA VERSION!!! **
Use at your own risk. You may damage your PC and/or VMU. Not for the
faint of heart. I'm assuming that the VMU isn't 5 volt compatible, so
make sure you take precautions to not over-drive it. This is especially
important on its *outputs* because PCs tend to pull up their inputs to
+5 volts (even though they sometimes drive at 3.3v). On my parallel port
a 4.7K ohm resistor to ground kept the voltage to 3 volts; your parallel
port may be different. Test it! I can't be responsible for any damage!
You've been warned!

Future improvements:

 - loops are used for timing. This probably is OK (the VMU seemed
   lax), but not the greatest for compatibility. This was tested 
   on a 366 MHz Pentium II Dell Inspiron 7000 (it's a laptop so it
   has only a 66 MHz bus).
 - verbose mode will almost certainly mess everything up. I didn't
   try it.
 - I messed up Soren's nice code. I've got to clean it up.
 - There is a possibility of making this interrupt-driven. (I selected
   the pinout based on this). It's not implemented. I hope the clock
   polarity is correct.

-------------------------------------------------------------------------
This driver eliminates the need for Soeren's Atmel circuit, but
not every parallel port will be compatible.

There are 14 pins on the VMU. They are actually labeled on its internal
PC board, so I'll keep the same convention. Looking from the back of
the unit with the connector at the top, the pins are numbered (from 
left to right) 1 through 14. Pins 1 through 7 are on one half of the
connector and pins 8 through 14 are on the other half.

Only these VMU pins are used in this application:
 
    as seen by VMU
          |
          v
VMU Pin In/Out VMU Signal Name 
------- ------ --------------- 
   3      O    Clock out
   4      I    Delayed data In
   5      O    Data out
   6      I    Handshake
 7,8      -    Ground
  10      I    Data in 
  11      O    Delayed Data out
  12      I    Clock in
   


LPT Pin In/Out LPT Signal Name      VMU Pin  VMU Name               #define
------- ------ ---------------      -------  ---------------------  -----------
   1     I/O   -STROBE         
   2     I/O   Data Bit 0              12     PC2VMU clock          OUT_CLK     
   3     I/O   Data Bit 1              10     PC2VMU data           OUT_DATA
   4     I/O   Data Bit 2               4     PC2VMU delayed data   OUT_DLYDATA
   5     I/O   Data Bit 3               6     PC2VMU handshake      OUT_HSK
   6     I/O   Data Bit 4      
   7     I/O   Data Bit 5      
   8     I/O   Data Bit 6      
   9     I/O   Data Bit 7      
  10      I     -ACK                    3     VMU2PC clock          IN_CLK
  11      I     BUSY                    5     VMU2PC data           IN_DATA
  12      I     PE                      
  13      I     SLCT                   11     VMU2PC delayed data   IN_DLYDATA
  14      O     -AUTO FEED XT           
  15      I     -ERROR
  16      O     -INIT                   
  17      O     -SLCT IN
18-25    ---    Ground                7,8     Ground
                                      1-14    Make sure these are connected so that
                                              the VMU's drivers are enabled.

The parallel port is pinned out as follows:

     1 1 1 1
     3 2 1 0 9 8 7 6 5 4 3 2 1
  .-----------------------------.
  \  o o o o o o o o o o o o o  /
   \  o o o o o o o o o o o o  /
    \-------------------------/
      2 2 2 2 2 2 1 1 1 1 1 1
      5 4 3 2 1 0 9 8 7 6 5 4




base  Data Latch (read/write)

base+1 Read-only: Printer Status
       +7-6-5-4-3-2-1-0+
            0 0 0
       +---------------+ bit
            +-------. 3: -ERROR (pin 15) 0=printer signals an error
           +---------. 4: +SLCT (pin 13)  1=printer is selected
          +-----------. 5: +PE (pin 12)    1=out of paper
         +-------------. 6: -ACK (pin 10)   0=ready for next character
        +---------------. 7: -BUSY (pin 11)  0=busy or off-line or error

base+2 Read/Write: Printer Controls
       +7-6-5-4-3-2-1-0+
       0 0      
       +---------------+ bit
                 +-. 0: +Strobe (pin 1)         1 when sending byte
                +---. 1: +AUTO LineFeed (pin 14) 1 causes LF after CR
               +-----. 2: -INIT (pin 16)          0 resets the printer
              +-------. 3: +SLCT IN (pin 17)       1 selects the printer
             +---------. 4: +IRQ Enable (hdwr INT when -ACK goes false)
                                         LPT1-.IRQ 7 (INT 0fH)
                                         LPT2-.IRQ 5 (INT 0dH)
            +-----------. 5: direction (PS/2 and later ATs)
                             0=write data to printer or device
                             1=read data from device


********************************************************************/


#define OUT_CLK     1
#define OUT_DATA    2
#define OUT_DLYDATA 4
#define OUT_HSK     8

#define IN_CLK     0x40
#define IN_DATA    0x80
#define IN_DLYDATA 0x10
 /* some of the bits are inverted by the parallel port controller.
    Use this XOR value to make all bits +3.3v = true = 1 */
#define IN_XOR     (0xc8 ^ IN_CLK)
/* not sure why clock needs to be inverted; the above docs
   make it look like it shouldn't need to be. */

/*******************************************************************
Copyright (C) 2000 by John Maushammer
http://www.tidalwave.net/~maushammer/vmu.html

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

You can get the latest Version from http://soeren.infoserv.de
********************************************************************/

/* #include <termios.h> */
#include <stdio.h>
/* #include <unistd.h> */
#include <fcntl.h>
/* #include <sys/time.h> */
#include <string.h>

#include "lowlevel_dosparallel.h"
#include "misc.h"


/* substitute for the library function for outp.
 * It's not a part of the default DEBUG library... go figure!
 */
#if 0
void outp(int short port, unsigned char data)
{
  __asm mov DX, port
  __asm mov AL, data
  __asm out DX,AL
};
#endif

/* low level delay. Uses bios function int 15 86 
 * CX,DX interval in microseconds (1,000,000ths of a second)
 *    CX is high word, DX is low word
 * 
 * It appears that the VMU can accept bit times that are
 * much quicker than this (but that doesn't mean it can accept
 * byte intervals that quick -- it still has to process each
 * byte using an interrupt service routine). I accidently tried
 * 8uS (the int 15 routine didn't work) and it was ok. I also
 * accidently tried 1.5uS and 2.3uS and those failed miserably.
 */
void wait64usec(void)
{
#if 0
  __asm mov DX, 0
  __asm mov CX, 64
  __asm mov AH, 86H
  __asm int 15
#else
  int i;
  for (i=0; i<10766; i++)
   _asm nop
#endif
}


/* open and initialise the serial port "serial" described by "serialname",
   return 0 on error, 1 on success */
int openserial(char *serialname, serial_t *serial)
{ 
  int short base;

  const int lptbases[3]={0x378, 0x278, 0x3bc};

  /* first, parse the name and get a base address */
  /* The BIOS stores LPT base port addresses at 40:0008, 40:000A, 40:000C, 40:000E */
  /* but we don't use that info because windows doesn't like us reading those addrs */
  if (strnicmp (serialname, "lpt", 3) == 0)  // if starts with LPT... (colon optional)
  {
     if ((serialname[3] >= '1') && (serialname[3] <= '3'))
     {
        base = lptbases[serialname[3]-'1'];               
     }
     else
     {  printf ("Unknown parallel port: %s\nTry a hex base address if necessary (ie. 0x378)\n", serialname);
        return (0);
     }
  }
  else   /* if doesn't start with lpt */
  {
     if (sscanf (serialname, "%hi", &base) != 1)
     {  printf ("Unknown parallel port: %s\nTry a hex base address if necessary (ie. 0x378)\n", serialname);
        return (0);
     }
  }

  printf ("open \"%s\" @ $%03x\n", serialname, (int) base);
  serial->port      = base;
  serial->outshadow = 0;
  serial->firstread = 1;   /* indicate that next read will be first one */
  serial->logptr =0;

  /* start communications by raising HSK true */
  serial->outshadow |= OUT_HSK | OUT_CLK; 

  outp (serial->port, serial->outshadow);

  return 1;
};



int closeserial(serial_t *serial)
{ int i;

  printf ("close\n");

  /* lower HSK to indicate end-of-comms */
  serial->outshadow &= ~OUT_HSK;
  outp (serial->port, serial->outshadow);

  printf ("\nCommunications log: ");
  for (i=0; i<serial->logptr; i++)
      printf ("%c:%02x ", (serial->log[i] & 0x100) ? 'R' : 'S', 
                           serial->log[i] & 0x2FF);
  printf ("\n");

  return 1;
};


/* This routine emulates the Atmel controller and actually performs
   one of two functions.

   The first time it is called, it simply monitors the handshaking
   line (IN_DLYDATA), and when it is 0 return 0x55.

   read one byte from the serial port "serial", give up after "timeout" seconds,
   return -1 on error, -2 on timeout, the byte received cast to an int on success 
*/


int readserial(serial_t *serial, int timeout)
{ 
  int short in;  /* read from parallel port */
  static byte lastc=0x50;
  static byte c;
  int bit;

  if (timeout <= 0)
    timeout=10;
  timeout *= 1000000;   /* fudge it: convert seconds time-out to loop count timeout */

  if (serial->firstread)
  { 
    printf ("firstread-");
    serial->firstread = 0;   /* normal reading next */
//return 0x55;
    while (timeout-- > 0)
    {
       in=inp(serial->port + 1) ^ IN_XOR;
       if ((in & IN_DLYDATA) == 0)
       {
         printf("hsk,");
         return 0x55;   /* got handshake */
       }
    }
    printf("err,");
    return -2;    /* timeout */
  }

//  printf ("read-");

#if 0
  /* paranoid check (may be removed if you're not paranoid) This may
     indicate corrupted data. If we've already gotten the clock, our
     processor may not be able to keep up with the VMU. */
  in=inp(serial->port + 1) ^ IN_XOR;
  if (in & IN_CLK)
  {
    printf ("paranoid,");
    return -1;  /* error */
  }
#endif

//    _asm CLI     ;disable interrupts
  /* wait for and read the byte */
  c=0;
  for (bit=7; (bit >=0) && (timeout>0); bit--)
  {

#if 0
    serial->outshadow &= ~OUT_DLYDATA;  /* data and clock low */
    outp (serial->port, serial->outshadow);
    serial->outshadow |= OUT_DLYDATA;   /* data high */
    outp (serial->port, serial->outshadow);
    serial->outshadow &= ~OUT_DLYDATA;  /* data and clock low */
    outp (serial->port, serial->outshadow);
    serial->outshadow |= OUT_DLYDATA;   /* data high */
    outp (serial->port, serial->outshadow);
    serial->outshadow &= ~OUT_DLYDATA;  /* data and clock low */
    outp (serial->port, serial->outshadow);
    serial->outshadow |= OUT_DLYDATA;   /* data high */
    outp (serial->port, serial->outshadow);
    serial->outshadow &= ~OUT_DLYDATA;  /* data and clock low */
    outp (serial->port, serial->outshadow);
    serial->outshadow |= OUT_DLYDATA;   /* data high */
    outp (serial->port, serial->outshadow);
    serial->outshadow &= ~OUT_DLYDATA;  /* data and clock low */
    outp (serial->port, serial->outshadow);
#endif

    /* wait for low clock (wait here while high) */
    do
    {
       in=inp(serial->port + 1) ^ IN_XOR;
    }
    while ( (--timeout>0) && ((in & IN_CLK) != 0) );

#if 0
    serial->outshadow &= ~OUT_DLYDATA;  /* data and clock low */
    outp (serial->port, serial->outshadow);
    serial->outshadow |= OUT_DLYDATA;   /* data high */
    outp (serial->port, serial->outshadow);
    serial->outshadow &= ~OUT_DLYDATA;  /* data and clock low */
    outp (serial->port, serial->outshadow);
    serial->outshadow |= OUT_DLYDATA;   /* data high */
    outp (serial->port, serial->outshadow);
    serial->outshadow &= ~OUT_DLYDATA;  /* data and clock low */
    outp (serial->port, serial->outshadow);
#endif

    /* setup previous data bit */
    if (lastc & (1 << bit))
       serial->outshadow |= OUT_DLYDATA;   /* data high */
    else
       serial->outshadow &= ~OUT_DLYDATA;  /* data and clock low */
    outp (serial->port, serial->outshadow);

    /* wait for rising clock */
    do
    {
       in=inp(serial->port + 1) ^ IN_XOR;
    }
    while ( (--timeout>0) && ((in & IN_CLK) == 0) );

    /* roll the data bit into the byte (msb first) */
    c = (c << 1) | ((in & IN_DATA) ? 1 : 0);

  }
//    _asm STI      ;enable interrupts again

  lastc=c;   /* remember this byte for the delayed version */

  if (timeout>0)
  { // printf("%02x,",c);
    serial->log[serial->logptr++] = 0x100 | c;
    return c;  /* got data in time */
  }
  else
  {
    printf ("only got %d bits\n", 6-bit);
    return -2;    /* timeout */
  }
};

/* The data lines for VMU-VMU communication seem to contain the same data,
   but one data line is an 8-bits delayed version of the other. I'm not sure
   why this is. Data is transferred most-signifcant-bit first.  The two
   clocks (one for each pair of data lines) are about 7812.5 Hz -- 64uS low
   and 64 uS high -- and don't run continuously. 
 
   This implemenation does not create the byte-delayed signal.

   Algorithm: for each bit (MSB first)
                output data & clock low
                wait 64 uS
                raise clock
                wait 64 uS
              repeat

   The timing waits at least 64uS and may be a little slow. The protocol
   seems to be tolerant of slowness.
 */

int writeserial(serial_t *serial, byte c, int timeout)
{ 
  int bit;
//  printf ("write-%02x,", (int) ((unsigned char) c));
  serial->log[serial->logptr++] = c;

  /* inter-character delay */
  for (bit=0; bit<5; bit++)
    wait64usec();

  for (bit=7; bit>=0; bit--)   /* send msb first */
  {
     /* output data and clock low*/
     if (c & (1 << bit))
     {
        serial->outshadow |= OUT_DATA;      /* data high */
        serial->outshadow &= ~OUT_CLK;      /* clock low */
     }
     else
        serial->outshadow &= ~(OUT_DATA | OUT_CLK);  /* data and clock low */
     outp (serial->port, serial->outshadow);

     wait64usec();

     /* clock high */
     serial->outshadow |= OUT_CLK;          /* clock high */
     outp (serial->port, serial->outshadow);

     wait64usec();
  }

  return 0;
};





#if 0
/* open and initialise the serial port "serial" described by "serialname",
   return 0 on error, 1 on success */
int openserial(char *serialname, serial_t *serial)
{
	struct termios serialio;

	/* open file */
	*serial = open(serialname, O_RDWR | O_SYNC);
	/* test for error */
	if (*serial < 0) return 0;
	/* get terminal attributes */
	if (tcgetattr(*serial, &serialio) != 0) return 0;
	/* make sure nobody changes any of the bytes as they go in and out */
	cfmakeraw(&serialio);
	/* 8 bits, receive, ignore modem control (DCD), enable RTS/CTS, drop DTR on close */
	serialio.c_cflag |= CS8 | CREAD | CLOCAL | CRTSCTS | HUPCL;
	/* 19200 bps in and out */
	if (cfsetospeed(&serialio, B19200) != 0) return 0;
	if (cfsetispeed(&serialio, B19200) != 0) return 0;
	/* set new parameters now */
	if (tcsetattr(*serial, TCSANOW, &serialio) != 0) return 0;
	return 1; /* everything OK */
}

/* close serial port "serial",
   return 0 on error, 1 on success */
int closeserial(serial_t *serial)
{
	int i;

	/* close serial port */
	i = close(*serial);
	/* destroy descriptor */
	*serial = 0;
	/* deliver return code */
	if (i == -1) return 0; else return 1;
}

/* read one byte from the serial port "serial", give up after "timeout" seconds,
   return -1 on error, -2 on timeout, the byte received cast to an int on success */
int readserial(serial_t *serial, int timeout)
{
	int i;
	fd_set readfds;
	struct timeval timeval;
	byte c;

	/* fill in timeout */
	timeval.tv_sec = timeout;
	timeval.tv_usec = 0;
	/* fill in file descriptor */
	FD_ZERO(&readfds);
	FD_SET(*serial, &readfds);
	/* call select */
	i = select(*serial + 1, &readfds, NULL, NULL, &timeval);
	/* error in select */
	if (i == -1) return -1;
	else {
		/* check if really ok to read */
		if (FD_ISSET(*serial, &readfds)) {
			/* read byte */
			i = read(*serial, &c, 1);
			/* check for error */
			if (i == 1) return (int) c; /* read 1 byte */
			else return -1;
		}
		/* select returned, but nothing to read: timeout */
		else return -2;
	}
}

/* wite one byte c to the serial port "serial", give up after "timeout" seconds,
   return -1 on error, -2 on timeout, 0 on success */
int writeserial(serial_t *serial, byte c, int timeout)
{
	int result;
	fd_set writefds;
	struct timeval timeval;

	/* fill in timeout */
	timeval.tv_sec = timeout;
	timeval.tv_usec = 0;
	/* fill in file descriptor */
	FD_ZERO(&writefds);
	FD_SET(*serial, &writefds);
	/* call select */
	result = select(*serial + 1, NULL, &writefds, NULL, &timeval);
	/* error in select */
	if (result == -1) return -1;
	else {
		/* check if really ok to write */
		if (FD_ISSET(*serial, &writefds)) {
			/* write byte */
			result = write(*serial, &c, 1);
			/* check for error */
			if (result == 1) return 0; /* 1 byte written, ok */
			else return -1;
		}
		/* select returned, but can't write: timeout */
		else return -2;
	}
}
#endif