/* send a file to SEGA Dreamcast VMU using my interface */
/* currently working on Linux and AmigaOS */

/* look at unknown: for things not yet implemented */

/* $Id: sendfile.c,v 1.13 2000/03/23 20:15:13 sgust Exp $ */

/*******************************************************************
Copyright (C) 2000 by Soeren Gust, sgust@ithh.infoserv.de

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

You can get the latest Version from http://soeren.infoserv.de
********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>

#ifndef _WIN32
#include <sys/time.h>
#include <getopt.h>
#else
#include "getopt.h"
#include <windows.h>
#include <winbase.h>
#define sleep Sleep
#include <time.h>
#endif

#include <signal.h>
#ifdef HAVE_UNISTD_H
/* no unistd.h on Amiga, sleep is in stdlib */
#include <unistd.h>
#endif

#include "lowlevel.h"
#include "misc.h"

/* parameters */
int verbose = 0;
int type = 0xcc;
char *vmuname = NULL;
char *serialname = DEFAULTDEV; 

/* global variable */
serial_t vmu; /* = (serial_t) 0; * filehandle for serial port */
byte *data = NULL; /* pointer to memory-block containing file data */

/* convert integer to BCD */
byte inttobcd(int i)
{
	return (i % 10) + (i / 10) * 16;
}

/* handle all errors: print message and abort */
void errorhandler(int result)
{
	/* no error: just return */
	if (result == SERTOVM_OK) return;
	/* something went wrong, we abort program, so close serial port */
/*	if (vmu != (serial_t) 0) closeserial(&vmu); */
    closeserial(&vmu);
	/* free memory, if allocated */
	if (data != NULL) free(data);
	switch(result) {
		case SERTOVM_TIMEOUT:
			if (verbose) printf("\n");
			fprintf(stderr, "Error: Timeout\n");
			exit(1);
		case SERTOVM_ILLEGALREQ:
			if (verbose) printf("\n");
			fprintf(stderr, "Error: Illegal request, please send bug-report\n");
			exit(1);
		case SERTOVM_ABORT:
			if (verbose) printf("\n");
			fprintf(stderr, "Error: VM wishes to abort\n");
			exit(1);
		case SERTOVM_PROTOCOLERROR:
			if (verbose) printf("\n");
			fprintf(stderr, "Error: Protocol error\n");
			exit(1);
		case SERTOVM_ERRNO:
			if (verbose) printf("\n");
			perror("Error");
			exit(1);
		case SERTOVM_SHORT:
			if (verbose) printf("\n");
			fprintf(stderr, "Error: Block too short, aborted\n");
			exit(1);
		case SERTOVM_TYPE:
			if (verbose) printf("\n");
			fprintf(stderr, "Error: Wrong block, aborted\n");
			exit(1);
		case SERTOVM_INTERNAL:
			if (verbose) printf("\n");
			fprintf(stderr, "Error: Internal error, please send bug-report\n");
			exit(1);
		default:
			if (verbose) printf("\n");
			fprintf(stderr, "Error: Unknown error, please send bug-report\n");
			exit(1);
	}
}

int main(int argc, char *argv[])
{
	FILE *file;
	byte block[12];
	struct stat status;
	int size, i, result;
	extern char *optarg;
	struct tm *time;

	/* parse parameters */
	while((i = getopt(argc, argv, "d:hn:t:v")) != -1) {
		switch(i) {
			case 'd':
				serialname = optarg;
				break;
			case 'h':
				fprintf(stderr, "Usage: %s [OPTIONS] FILE\n", argv[0]);
				fprintf(stderr, "Send a FILE to a Sega Dreamcast VM\n");
				fprintf(stderr, "Version 1.1\n");
				fprintf(stderr, "\n");
				fprintf(stderr, "  -d device  set device where the VMU is connected\n");
				fprintf(stderr, "  -t type    filetype 0x33 = data, 0xcc = game\n");
				fprintf(stderr, "  -n name    filename inside VMU\n");
				fprintf(stderr, "  -v         verbose\n");
				fprintf(stderr, "  -h         This help text\n");
				exit(1);
			case 'v':
				verbose = 1;
				break;
			case 't':
				type = strtol(optarg, NULL, 0);
				break;
			case 'n':
				vmuname = optarg;
				break;
			case '?':
				exit(1);
			default:
				perror("Problem with getopt");
				exit(1);
		}
	}
	/* begin parameter checks */
	/* last argument is filename */
	if (argc != optind + 1) {
		fprintf(stderr, "Error: Missing file parameter\n");
		exit(1);
	}
	/* only allow data/game */
	if ((type != 0xcc) && (type != 0x33)) {
		fprintf(stderr, "Error: Type can't be %02x\n", type);
		exit(1);
	}
	/* internal name = filename if not set */
	if (vmuname == NULL) vmuname = argv[argc - 1];
	/* end parameter check */

	file = fopen(argv[argc - 1], "rb");
	if (file == NULL) {
		perror("Can't open file");
		closeserial(&vmu);
		exit(1);
	}
	/* may not be portable */
	if(stat(argv[argc - 1], &status) != 0) {
		perror("Can't stat file");
		closeserial(&vmu);
		exit(1);
	}
	size = status.st_size;
	if (verbose) printf("Size of file %d bytes, %d blocks\n", size, size / 512);
	if (size % 512) {
		fprintf(stderr, "Error: Size not dividable by 512\n");
		closeserial(&vmu);
		exit(1);
	}
	/* allocate memory for file */
	data = malloc(size);
	if (data == 0) {
		perror("Out of memory for file");
		closeserial(&vmu);
		exit(1);
	}
	/* read file contents */
	if (fread(data, size, 1, file) != 1) {
		perror("Error reading file");
		closeserial(&vmu);
		free(data);
		exit(1);
	}
	/* close file */
	if (fclose(file) != 0) {
		perror("Error closing file");
		closeserial(&vmu);
		free(data);
		exit(1);
	}

	result = openserial(serialname, &vmu);
	if (result != 1) {
		perror("Can't open serial port");
		exit(1);
	}
	/* VM ready */
	if (verbose) {
		printf("Waiting for VM ready ");
		fflush(stdout);
	}
vmu.log[vmu.logptr++] = 0x200;
	/* get VM ready byte */
	result = readserial(&vmu, 10);
	if (result == -1) {
		if (verbose) printf("\n");
		perror("Error getting VM ready");
		closeserial(&vmu);
		free(data);
		exit(1);
	}
	if (result == -2) {
		if (verbose) printf("\n");
		fprintf(stderr, "Error: Timeout getting VM ready\n");
		closeserial(&vmu);
		free(data);
		exit(1);
	}
	if (result != 0x55) {
		if (verbose) printf("\n");
		fprintf(stderr, "Error: Protocol error in VM ready\n");
		closeserial(&vmu);
		free(data);
		exit(1);
	}
	if (verbose) printf("OK\n");
vmu.log[vmu.logptr++] = 0x201;

#if 0
doesn't seem to be working
	sleep(1); /* it takes second for the VMU to get ready for the first block */
#else
  for (i=0; i<168000000; i++)   /* 1 second */
    _asm nop;
#endif

	/* start block */
	if (verbose) {
		printf("Sending start block ");
		fflush(stdout);
	}
	/* build type 2 block */
	block[0] = type;
	block[1] = size / 512;
	/* send start block */
	result = sendblock(&vmu, 2, 2, block, 5);
vmu.log[vmu.logptr++] = 0x202;
	if (result == SERTOVM_ABORT)
	{
		if (verbose) printf("\n");
		fprintf(stderr, "Error: Not enough free memory on VM\n");
		closeserial(&vmu);
		free(data);
		exit(1);
	}
	/* handle other errors */
	errorhandler(result);
	if (verbose) printf("OK\n");

	/* name block */
	if (verbose) {
		printf("Sending name block ");
		fflush(stdout);
	}
	/* compose name header */
	i = 0;
	while((i < 12) && (vmuname[i] != 0)) { /* copy name */
		block[i] = vmuname[i];
		i++;
	}
	while(i < 12) { /* fill rest of name with spaces */
		block[i++] = ' ';
	}
	/* send name block */
	result = sendblock(&vmu, 12, 6, block, 5);
	if (result == SERTOVM_ABORT)
	{
		if (verbose) printf("\n");
		fprintf(stderr, "Error: Name already in use on VM\n");
		closeserial(&vmu);
		free(data);
		exit(1);
	}
	/* handle other errors */
	errorhandler(result);
	if (verbose) printf("OK\n");

	/* send file contents */
	for(i = 0; i < size / 128; i++) {
		/* data block */
		if (verbose) {
			printf("Sending data %03d/%03d ", i + 1, size / 128);
			fflush(stdout);
		}
		result = sendblock(&vmu, 128, 4, &data[i * 128], 5);
		/* handle errors */
		errorhandler(result);

		/* flash block */
		if (verbose) {
			putchar(13); /* portable ? */
			fflush(stdout);
		}
		block[0] = 0; /* unknown: what does this mean? */
		result = sendblock(&vmu, 1, 8, block, 5);
		/* handle errors */
		errorhandler(result);
	}

	/* directory block */
	if (verbose) {
		printf("\nSending directory block ");
		fflush(stdout);
	}
	/* compose directory block */
	time = localtime(&status.st_mtime);
	block[0] = inttobcd((time->tm_year + 1900) / 100);
	block[1] = inttobcd((time->tm_year + 1900) % 100);
	block[2] = inttobcd(time->tm_mon + 1);
	block[3] = inttobcd(time->tm_mday);
	block[4] = inttobcd(time->tm_hour);
	block[5] = inttobcd(time->tm_min);
	block[6] = inttobcd(time->tm_sec);
	block[7] = time->tm_wday; /* always 0-6, no conversion necessary */
	block[8] = size / 512; /* size low byte */
	block[9] = 0; /* size high byte */
	block[10] = (type == 0xcc)?1:0; /* header low byte, game:1, else 0 */
	block[11] = 0; /* header high byte */
	/* send directory block */
	result = sendblock(&vmu, 12, 3, block, 5);
	/* handle errors */
	errorhandler(result);
	if (verbose) printf("OK\n");

	/* last flash block */
	if (verbose) {
		printf("Sending last flash block ");
		fflush(stdout);
	}
	block[0] = 0; /* unknown: what does this mean? */
	result = sendblock(&vmu, 1, 8, block, 5);
	/* handle errors */
	errorhandler(result);
	if (verbose) printf("OK\n");

	/* file succesfully transfered  */
	closeserial(&vmu);
	free(data);
	exit(0);
	return(0);
}
