/* SerialToVM: Interface between RS-232 and SEGA Dreamcast VM */
/* Implements several systemdependant low-lovel routines */
/* This file is for AmigaOS */

/* $Id: lowlevel_amiga.c,v 1.3 2000/03/17 14:05:32 sgust Exp $ */

/*******************************************************************
Copyright (C) 2000 by Soeren Gust, sgust@ithh.infoserv.de

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

You can get the latest Version from http://soeren.infoserv.de
********************************************************************/

/* several amiga specific includes */
#include <exec/types.h>
#include <exec/memory.h>
#include <exec/semaphores.h>
#include <exec/io.h>
#include <exec/devices.h>
#include <exec/errors.h>
#include <devices/serial.h>
#include <devices/timer.h>
#include <dos/dos.h> /* just for SIGBREAKF_CTRL_C */

#include <clib/exec_protos.h>
#include <clib/alib_protos.h>

/* libc includes */
#include <strings.h>
#include <stdlib.h>
#include <signal.h>

#include "lowlevel_amiga.h"
#include "misc.h"

static struct amigaserial gs;

/* we need this to close the serial.device when the user presses ^C */
void breakhandler(int signo)
{
	serial_t *serial;

	*serial = &gs;
	closeserial(serial);
	exit(5);
}

/* open and initialise the serial port "serial" described by "serialname",
   return 0 on error, 1 on success */
int openserial(char *serialname, serial_t *serial)
{
	char *p = NULL, devname[101];
	int unit = 0;

	/* init amigaserial structure */
	gs.serialport = NULL;
	gs.serialio = NULL;
	gs.serialopen = -1;
	gs.timerport = NULL;
	gs.timerio = NULL;
	gs.timeropen = -1;
	*serial = &gs;
	/* start signal handler */
	signal(SIGINT, breakhandler);
	/* port for serial */
	gs.serialport = CreatePort(0, 0);
	if (gs.serialport == NULL) {
		gs.serialio = NULL;
		return 0;
	}
	/* port for timer */
	gs.timerport = CreatePort(0, 0);
	if (gs.timerport == NULL) {
		closeserial(serial);
		return 0;
	}
	/* IO for serial */
	gs.serialio = (struct IOExtSer *) CreateExtIO(gs.serialport, sizeof(struct IOExtSer));
	if (gs.serialio == NULL) {
		closeserial(serial);
		return 0;
	}
	/* IO for timer */
	gs.timerio = (struct timerequest *) CreateExtIO(gs.timerport, sizeof(struct timerequest));
	if (gs.timerio == NULL) {
		closeserial(serial);
		return 0;
	}
	/* set serial flags */
	gs.serialio->io_SerFlags=SERF_7WIRE | SERF_XDISABLED;
	/* get name/unit */
	strncpy(devname, serialname, 100);
	devname[101] = 0;
	p = strchr(devname, '/');
	if (p != NULL) {
		unit = atoi(p + 1);
		*p = 0;
	}
	/* open serial */
	gs.serialopen = OpenDevice(devname, unit, (struct IORequest *) gs.serialio, 0);
	if (gs.serialopen != 0) {
		closeserial(serial);
		return 0;
	}
	/* open timer */
	gs.timeropen = OpenDevice(TIMERNAME, UNIT_VBLANK, (struct IORequest *) gs.timerio, 0);
	if (gs.timeropen != 0) {
		closeserial(serial);
		return 0;
	}
	/* set serial parameter */
	gs.serialio->io_Baud = 19200;
	gs.serialio->io_ReadLen = gs.serialio->io_WriteLen = 8;
	gs.serialio->io_StopBits = 1;
	gs.serialio->io_SerFlags=SERF_7WIRE | SERF_XDISABLED;
	gs.serialio->IOSer.io_Command = SDCMD_SETPARAMS;
	DoIO((struct IORequest *) gs.serialio);
	return 1; /* everything OK */
}

/* close serial port "serial",
   return 0 on error, 1 on success */
int closeserial(serial_t *serial)
{
	/* test for valid parameter */
	if (*serial != &gs) return 0;
	/* close serial device */
	if (gs.serialopen == 0) {
		CloseDevice((struct IORequest *) gs.serialio);
		gs.serialopen = -1;
	}
	/* delete serialio */
	if (gs.serialio != NULL) {
		DeleteExtIO((struct IORequest *) gs.serialio);
		gs.serialio = NULL;
	}
	/* remove serial port */
	if (gs.serialport != NULL) {
		DeletePort(gs.serialport);
		gs.serialport = NULL;
	}
	/* close timer device */
	if (gs.timeropen == 0) {
		CloseDevice((struct IORequest *) gs.timerio);
		gs.timeropen = -1;
	}
	/* delete timerequest */
	if (gs.timerio != NULL) {
		DeleteExtIO((struct IORequest *) gs.timerio);
		gs.timerio = NULL;
	}
	/* remove timer port */
	if (gs.timerport != NULL) {
		DeletePort(gs.timerport);
		gs.timerport = NULL;
	}
	/* success */
	return 1;
}

/* read one byte from the serial port "serial", give up after "timeout" seconds,
   return -1 on error, -2 on timeout, the byte received cast to an int on success */
/* Amiga: return -2 (timeout) when signal BREAK C arrives */
int readserial(serial_t *serial, int timeout)
{
	byte c;
	int serok; /* specific to DICE, should be struct IORequest * */
	ULONG waitbits;

	/* test for valid parameter */
	if (*serial != &gs) return -1;
	/* prepare serialio */
	gs.serialio->IOSer.io_Length = 1;
	gs.serialio->IOSer.io_Data = &c;
	gs.serialio->IOSer.io_Command = CMD_READ;
	/* fill in timeout */
	gs.timerio->tr_node.io_Command = TR_ADDREQUEST;
	gs.timerio->tr_time.tv_secs = timeout;
	gs.timerio->tr_time.tv_micro = 0;
	/* send serial and time requests */
	SendIO((struct IORequest *) gs.serialio);
	SendIO((struct IORequest *) gs.timerio);
	/* now wait for any of the requests to return */
	while(!CheckIO((struct IORequest *) gs.serialio) && !CheckIO((struct IORequest *) gs.timerio)) {
		waitbits = Wait(SIGBREAKF_CTRL_C | (1 << gs.serialport->mp_SigBit) | (1 << gs.timerport->mp_SigBit));
		/* abort serial when break, gets returned as timeout later */
		if (waitbits & SIGBREAKF_CTRL_C) AbortIO((struct IORequest *) gs.serialio);
	}
	/* check serial io */
	serok = CheckIO((struct IORequest *) gs.serialio);
	/* abort if not ready */
	if (serok == NULL) AbortIO((struct IORequest *) gs.serialio);
	/* abort timer */
	AbortIO((struct IORequest *) gs.timerio);
	/* wait for requests to return */
	WaitIO((struct IORequest *) gs.serialio);
	WaitIO((struct IORequest *) gs.timerio);
	if ((serok != NULL) && (gs.serialio->IOSer.io_Actual == 1)) return (int) c;
	if ((gs.serialio->IOSer.io_Error != 0) && (gs.serialio->IOSer.io_Error != IOERR_ABORTED)) return -1;
	return -2;
}

/* wite one byte c to the serial port "serial", give up after "timeout" seconds,
   return -1 on error, -2 on timeout, 0 on success */
/* Amiga: return -2 (timeout) when signal BREAK C arrives */
int writeserial(serial_t *serial, byte c, int timeout)
{
	int serok; /* specific to DICE, should be struct IORequest * */
	ULONG waitbits;

	/* test for valid parameter */
	if (*serial != &gs) return -1;
	/* prepare serialio */
	gs.serialio->IOSer.io_Length = 1;
	gs.serialio->IOSer.io_Data = &c;
	gs.serialio->IOSer.io_Command = CMD_WRITE;
	/* fill in timeout */
	gs.timerio->tr_node.io_Command = TR_ADDREQUEST;
	gs.timerio->tr_time.tv_secs = timeout;
	gs.timerio->tr_time.tv_micro = 0;
	/* send serial and time requests */
	SendIO((struct IORequest *) gs.serialio);
	SendIO((struct IORequest *) gs.timerio);
	/* now wait for any of the requests to return */
	while(!CheckIO((struct IORequest *) gs.serialio) && !CheckIO((struct IORequest *) gs.timerio)) {
		waitbits = Wait(SIGBREAKF_CTRL_C | (1 << gs.serialport->mp_SigBit) | (1 << gs.timerport->mp_SigBit));
		/* abort serial when break, gets returned as timeout later */
		if (waitbits & SIGBREAKF_CTRL_C) AbortIO((struct IORequest *) gs.serialio);
	}
	/* check serial io */
	serok = CheckIO((struct IORequest *) gs.serialio);
	/* abort if not ready */
	if (serok == NULL) AbortIO((struct IORequest *) gs.serialio);
	/* abort timer */
	AbortIO((struct IORequest *) gs.timerio);
	/* wait for requests to return */
	WaitIO((struct IORequest *) gs.serialio);
	WaitIO((struct IORequest *) gs.timerio);
	if ((serok != NULL) && (gs.serialio->IOSer.io_Actual == 1)) return 0;
	if ((gs.serialio->IOSer.io_Error != 0) && (gs.serialio->IOSer.io_Error != IOERR_ABORTED)) return -1;
	return -2;
}
