/* SerialToVM: Interface between RS-232 and SEGA Dreamcast VM */
/* Implements several systemindependant mid-level routines handling blocks */

/* $Id: midlevel.c,v 1.4 2000/03/07 12:53:44 sgust Exp $ */

/*******************************************************************
Copyright (C) 2000 by Soeren Gust, sgust@ithh.infoserv.de

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

You can get the latest Version from http://soeren.infoserv.de
********************************************************************/

#include <stdio.h>
#include "lowlevel.h"
#include "misc.h"

/* sends a VM block to "serial" of "length" bytes from "buffer" waiting
   max "timeout" seconds on every byte.
   returns any of the SERTOVM_ codes */
int sendblock(serial_t *serial, int length, byte type, byte *buffer, int timeout)
{
	int i, result;
	byte start; /* block header byte */
	byte end; /* block end marker */

	/* get block framing from length */
	if (length == 128) {
		start = 0x50;
		end = 0;
	}
	else if ((length >= 1) && (length <= 16)) {
		start = length + 0x0f;
		end = 0xff;
	}
	else return SERTOVM_ILLEGALREQ;
	/* send start byte */
	result = writeserial(serial, start, timeout);
	if (result == -2) return SERTOVM_TIMEOUT;
	if (result != 0) return SERTOVM_ERRNO;
	/* wait for ACK (0xe0) */
	result = readserial(serial, timeout);
	if (result == -2) return SERTOVM_TIMEOUT;
	if (result == -1) return SERTOVM_ERRNO;
	if (result != 0xe0) return SERTOVM_PROTOCOLERROR;
	/* send type */
	result = writeserial(serial, type, timeout);
	if (result == -2) return SERTOVM_TIMEOUT;
	if (result != 0) return SERTOVM_ERRNO;
	/* send data */
	for(i = 0; i < length; i++) {
		result = writeserial(serial, buffer[i], timeout);
		if (result == -2) return SERTOVM_TIMEOUT;
		if (result != 0) return SERTOVM_ERRNO;
	}
	/* send end byte */
	result = writeserial(serial, end, timeout);
	if (result == -2) return SERTOVM_TIMEOUT;
	if (result != 0) return SERTOVM_ERRNO;
	/* wait for ACK (0x0c) */
	result = readserial(serial, timeout);
	if (result == -2) return SERTOVM_TIMEOUT;
	if (result == -1) return SERTOVM_ERRNO;
	if (result == 0x0a) return SERTOVM_ABORT;
	if (result != 0x0c) return SERTOVM_PROTOCOLERROR;
	return SERTOVM_OK;
}

/* get a VM block from "serial" of "length" bytes to "buffer" waiting
   max "timeout" seconds on every byte expecting the block to be "type" 
   or any (-1). If ack is 1 the block will be acknowledged.
   returns any of the SERTOVM_ codes and the type of the block in type */
int receiveblock(serial_t *serial, int length, int *type, byte *buffer, int timeout, int ack)
{
	int i, result, reallength, realtype;
	byte last;

	reallength = 0;
	/* wait for start  */
	result = readserial(serial, timeout);
	if (result == -2) return SERTOVM_TIMEOUT;
	if (result == -1) return SERTOVM_ERRNO;
	if (result == 0x50) reallength = 128;
	else if ((result >= 0x10) && (result <= 0x1f)) reallength = result - 0x0f;
	/* real VMs behave different: they just wait for a valid block start */
	else return SERTOVM_PROTOCOLERROR;
	/* send ACK (0xe0) */
	result = writeserial(serial, 0xe0, timeout);
	if (result == -2) return SERTOVM_TIMEOUT;
	if (result != 0) return SERTOVM_ERRNO;
	/* get type */
	result = readserial(serial, timeout);
	if (result == -2) return SERTOVM_TIMEOUT;
	if (result == -1) return SERTOVM_ERRNO;
	realtype = (char) result;
	/* get data */
	for(i = 0; i < reallength; i++) {
		result = readserial(serial, timeout);
		if (result == -2) return SERTOVM_TIMEOUT;
		if (result == -1) return SERTOVM_ERRNO;
		/* only store byte if space left */
		if (i < length) buffer[i] = (char) result;
	}
	/* get end */
	result = readserial(serial, timeout);
	if (result == -2) return SERTOVM_TIMEOUT;
	if (result == -1) return SERTOVM_ERRNO;
	last = (char) result;
	/* wrong block type, abort */
	if ((*type != realtype) && (*type != -1)) {
		if (ack == 1) {
			result = writeserial(serial, 0x0a, timeout);
			if (result == -2) return SERTOVM_TIMEOUT;
			if (result != 0) return SERTOVM_ERRNO;
		}
		return SERTOVM_TYPE;
	}
	/* wrong type, block too short: send abort */
	if (reallength < length) {
		if (ack == 1) {
			result = writeserial(serial, 0x0a, timeout);
			if (result == -2) return SERTOVM_TIMEOUT;
			if (result != 0) return SERTOVM_ERRNO;
		}
		return SERTOVM_SHORT;
	}
	/* wrong last byte, abort nicely */
	else if ((last != 0) && (last != 0xff)) {
		result = writeserial(serial, 0x0a, timeout);
		if (result == -2) return SERTOVM_TIMEOUT;
		if (result != 0) return SERTOVM_ERRNO;
		else return SERTOVM_PROTOCOLERROR;
	}
	/* all ok, send continue */
	else if (ack == 1) {
		result = writeserial(serial, 0x0c, timeout);
		if (result == -2) return SERTOVM_TIMEOUT;
		if (result != 0) return SERTOVM_ERRNO;
		return SERTOVM_OK;
	}
	else return SERTOVM_OK;
	return SERTOVM_INTERNAL; /* just in case something is wrong in the code */
}
