/* receive a file from SEGA Dreamcast VMU using my interface */
/* currently working on Linux and AmigaOS */

/* look at unknown: for things not yet implemented */

/* $Id: receivefile.c,v 1.4 2000/03/23 20:15:13 sgust Exp $ */

/*******************************************************************
Copyright (C) 2000 by Soeren Gust, sgust@ithh.infoserv.de

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

You can get the latest Version from http://soeren.infoserv.de
********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <getopt.h>
#include <signal.h>
#ifdef HAVE_UTIME_H
#include <utime.h>
#endif
#ifdef HAVE_UTIME_AMIGA_H
#include <utime_amiga.h>
#endif
#ifdef HAVE_UNISTD_H
/* no unistd.h on Amiga, sleep is in stdlib */
#include <unistd.h>
#endif

#include "lowlevel.h"
#include "misc.h"

/* parameters */
int verbose = 0;
char *serialname = DEFAULTDEV;

/* global variable */
serial_t vmu = (serial_t) 0; /* filehandle for serial port */
byte *data = NULL; /* pointer to memory-block containing file data */

/* convert BCD to integer */
int bcdtoint(byte i)
{
	return (i % 16) + (i / 16) * 10;
}

/* handle all errors: print message and abort */
void errorhandler(int result)
{
	/* no error: just return */
	if (result == SERTOVM_OK) return;
	/* something went wrong, we abort program, so close serial port */
	if (vmu != (serial_t) 0) closeserial(&vmu);
	/* free memory, if allocated */
	if (data != NULL) free(data);
	switch(result) {
		case SERTOVM_TIMEOUT:
			if (verbose) printf("\n");
			fprintf(stderr, "Error: Timeout\n");
			exit(1);
		case SERTOVM_ILLEGALREQ:
			if (verbose) printf("\n");
			fprintf(stderr, "Error: Illegal request, please send bug-report\n");
			exit(1);
		case SERTOVM_ABORT:
			if (verbose) printf("\n");
			fprintf(stderr, "Error: VM wishes to abort\n");
			exit(1);
		case SERTOVM_PROTOCOLERROR:
			if (verbose) printf("\n");
			fprintf(stderr, "Error: Protocol error\n");
			exit(1);
		case SERTOVM_ERRNO:
			if (verbose) printf("\n");
			perror("Error");
			exit(1);
		case SERTOVM_SHORT:
			if (verbose) printf("\n");
			fprintf(stderr, "Error: Block too short, aborted\n");
			exit(1);
		case SERTOVM_TYPE:
			if (verbose) printf("\n");
			fprintf(stderr, "Error: Wrong block, aborted\n");
			exit(1);
		case SERTOVM_INTERNAL:
			if (verbose) printf("\n");
			fprintf(stderr, "Error: Internal error, please send bug-report\n");
			exit(1);
		default:
			if (verbose) printf("\n");
			fprintf(stderr, "Error: Unknown error, please send bug-report\n");
			exit(1);
	}
}

int main(int argc, char *argv[])
{
	FILE *file;
	byte block[12], filename[13], c, *myname = filename;
	struct stat status;
	int size, type, i, result, blocktype;
	extern char *optarg;
	struct tm time;
	struct utimbuf utim;

	/* parse parameters */
	while((i = getopt(argc, argv, "d:hn:v")) != -1) {
		switch(i) {
			case 'd':
				serialname = optarg;
				break;
			case 'h':
				fprintf(stderr, "Usage: %s [OPTIONS]\n", argv[0]);
				fprintf(stderr, "Receive a file from a Sega Dreamcast VM\n");
				fprintf(stderr, "Version 1.1\n");
				fprintf(stderr, "\n");
				fprintf(stderr, "  -d device  set device where the VMU is connected\n");
				fprintf(stderr, "  -n name    filename to use\n");
				fprintf(stderr, "  -v         verbose\n");
				fprintf(stderr, "  -h         This help text\n");
				exit(1);
			case 'v':
				verbose = 1;
				break;
			case 'n':
				myname = optarg;
				break;
			case '?':
				exit(1);
			default:
				perror("Problem with getopt");
				exit(1);
		}
	}
	/* begin parameter checks */
	/* no argument */
	if (argc != optind) {
		fprintf(stderr, "Error: Additional file parameter\n");
		exit(1);
	}
	/* end parameter check */
	/* open serial port */
	result = openserial(serialname, &vmu);
	if (result != 1) {
		perror("Can't open serial port");
		exit(1);
	}

	/* VM ready */
	if (verbose) {
		printf("Waiting for VM ready ");
		fflush(stdout);
	}
	/* get VM ready byte */
	result = readserial(&vmu, 10);
	if (result == -1) {
		if (verbose) printf("\n");
		perror("Error getting VM ready");
		closeserial(&vmu);
		exit(1);
	}
	if (result == -2) {
		if (verbose) printf("\n");
		fprintf(stderr, "Error: Timeout getting VM ready\n");
		closeserial(&vmu);
		exit(1);
	}
	if (result != 0x55) {
		if (verbose) printf("\n");
		fprintf(stderr, "Error: Protocol error in VM ready\n");
		closeserial(&vmu);
		exit(1);
	}
	if (verbose) printf("OK\n");

	sleep(1); /* seems necessary, doesn't work without */

	/* start block */
	if (verbose) {
		printf("Waiting for start block ");
		fflush(stdout);
	}
	/* expect type 2 block */
	blocktype = 2;
	result = receiveblock(&vmu, 2, &blocktype, block, 5, 0);
	/* handle errors */
	errorhandler(result);
	/* check parameters */
	type = block[0];
	size = block[1] * 512;
	if ((type != 0xcc) && (type != 0x33)) {
		if (verbose) printf("\n");
		fprintf(stderr, "Error: File type 0x%02x is neither Game nor Data\n", type);
		writeserial(&vmu, 0x0a, 5);
		closeserial(&vmu);
		exit(1);
	}
	/* allocate memory */
	data = malloc(size);
	if (data == NULL) {
		if (verbose) printf("\n");
		fprintf(stderr, "Error: Out of memory\n");
		writeserial(&vmu, 0x0a, 5);
		closeserial(&vmu);
		exit(1);
	}
	/* acknowledge block */
	result = writeserial(&vmu, 0x0c, 5);
	if (result == -1) {
		if (verbose) printf("\n");
		perror("Error in acknowledge start block");
		closeserial(&vmu);
		exit(1);
	}
	if (result == -2) {
		if (verbose) printf("\n");
		fprintf(stderr, "Error: Timeout in acknowledge start block\n");
		closeserial(&vmu);
		exit(1);
	}
	if (verbose) {
		printf("OK\nType: %s, size %d blocks\n", type == 0x33?"Data":"Game", size / 512);
		fflush(stdout);
	}
	
	/* name block */
	if (verbose) {
		printf("Waiting for name block ");
		fflush(stdout);
	}
	blocktype = 6;
	result = receiveblock(&vmu, 12, &blocktype, filename, 5, 0);
	filename[12] = 0;
	/* delete pending spaces */
	for(i = 11; i != 0; i--) if (filename[i] == ' ') filename[i] = 0;
	/* handle errors */
	errorhandler(result);
	/* look, if file already there */
	if (!stat(myname, &status)) {
		if (verbose) printf("\n");
		fprintf(stderr, "Error: File %s already exists\n", myname);
		writeserial(&vmu, 0x0a, 5);
		closeserial(&vmu);
		free(data);
		exit(1);
	}
	file = fopen(myname, "wb");
	if (file == NULL) {
		if (verbose) printf("\n");
		fprintf(stderr, "Error: Can't open file %s\n", myname);
		writeserial(&vmu, 0x0a, 5);
		closeserial(&vmu);
		free(data);
		exit(1);
	}
	/* acknowledge block */
	result = writeserial(&vmu, 0x0c, 5);
	if (result == -1) {
		if (verbose) printf("\n");
		perror("Error in acknowledge name block");
		closeserial(&vmu);
		free(data);
		exit(1);
	}
	if (result == -2) {
		if (verbose) printf("\n");
		fprintf(stderr, "Error: Timeout in acknowledge name block\n");
		closeserial(&vmu);
		free(data);
		exit(1);
	}
	if (verbose) {
		printf("OK\nName: %s\n", myname);
		fflush(stdout);
	}


	/* receive file contents */
	for(i = 0; i < size / 128; i++) {
		/* data block */
		if (verbose) {
			printf("Receiving data %03d/%03d ", i + 1, size / 128);
			fflush(stdout);
		}
		blocktype = 4;
		result = receiveblock(&vmu, 128, &blocktype, &data[i * 128], 5, 1);
		/* handle errors */
		errorhandler(result);

		/* flash block */
		if (verbose) {
			putchar(13); /* portable ? */
			fflush(stdout);
		}
		blocktype = 8;
		result = receiveblock(&vmu, 1, &blocktype, block, 5, 1);
		/* handle errors */
		errorhandler(result);
	}

	/* directory block */
	if (verbose) {
		printf("\nReceiving directory block ");
		fflush(stdout);
	}
	blocktype = 3;
	result = receiveblock(&vmu, 12, &blocktype, block, 5, 1);
	/* handle errors */
	errorhandler(result);
	if (verbose) printf("OK\n");

	/* last flash block */
	if (verbose) {
		printf("Receiving last flash block ");
		fflush(stdout);
	}
	blocktype = 8;
	result = receiveblock(&vmu, 1, &blocktype, &c, 5, 1);
	/* handle errors */
	errorhandler(result);
	if (verbose) printf("OK\n");

	/* file succesfully transferred, save now */
	if (fwrite(data, size, 1, file) != 1) {
		perror("Error writing file");
		closeserial(&vmu);
		free(data);
		exit(1);
	}
	fclose(file);
	/* touch file */
	time.tm_year = bcdtoint(block[1]) + 100 * bcdtoint(block[0]) - 1900;
	time.tm_mon = bcdtoint(block[2]) - 1;
	time.tm_mday = bcdtoint(block[3]);
	time.tm_hour = bcdtoint(block[4]);
	time.tm_min = bcdtoint(block[5]);
	time.tm_sec = bcdtoint(block[6]);
	time.tm_isdst = 0;
	utim.actime = utim.modtime = mktime(&time);
	utime(myname, &utim);
	/* close everything */
	closeserial(&vmu);
	free(data);
	exit(0);
	return(0);
}
